<?php

namespace Drupal\notification_system_dispatch_webpush\Plugin\NotificationSystemDispatcher;

use Drupal\Component\Render\FormattableMarkup;
use Drupal\Core\Asset\LibraryDiscoveryInterface;
use Drupal\Core\Config\Config;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Datetime\DateFormatterInterface;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\Mail\MailFormatHelper;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\Render\ElementInfoManagerInterface;
use Drupal\Core\Render\Markup;
use Drupal\Core\StreamWrapper\StreamWrapperManager;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\Core\Template\TwigEnvironment;
use Drupal\Core\Url;
use Drupal\file\Entity\File;
use Drupal\notification_system_dispatch\NotificationSystemDispatcherPluginBase;
use Drupal\notification_system_dispatch_webpush\WebPushClient;
use Drupal\user\UserInterface;
use Drupal\web_push_api\Component\WebPushData;
use Drupal\web_push_api\Component\WebPushNotification;
use Html2Text\Html2Text;
use Psr\Log\LoggerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use function file_create_url;

/**
 * Plugin implementation of the notification_system_dispatcher.
 *
 * @NotificationSystemDispatcher(
 *   id = "webpush",
 *   label = @Translation("Push-Notification"),
 *   description = @Translation("Send notifications via web push.")
 * )
 */
class WebpushDispatcher extends NotificationSystemDispatcherPluginBase implements ContainerFactoryPluginInterface {

  use StringTranslationTrait;

  /**
   * The TwigEnvironment.
   *
   * @var \Drupal\Core\Template\TwigEnvironment
   */
  protected TwigEnvironment $twig;


  /**
   * The DateFormatter.
   *
   * @var \Drupal\Core\Datetime\DateFormatterInterface
   */
  private DateFormatterInterface $dateFormatter;

  /**
   * The WebPushClient.
   *
   * @var \Drupal\notification_system_dispatch_webpush\WebPushClient
   */
  protected WebPushClient $webPushClient;

  /**
   * The FileSystem.
   *
   * @var \Drupal\Core\File\FileSystemInterface
   */
  protected FileSystemInterface $fileSystem;

  /**
   * The LibraryDiscovery.
   *
   * @var \Drupal\Core\Asset\LibraryDiscoveryInterface
   */
  protected LibraryDiscoveryInterface $libraryDiscovery;

  /**
   * The ElementInfoManager.
   *
   * @var \Drupal\Core\Render\ElementInfoManagerInterface
   */
  protected ElementInfoManagerInterface $elementInfoManager;

  /**
   * The logger.
   *
   * @var \Psr\Log\LoggerInterface
   */
  protected LoggerInterface $logger;

  /**
   * The configuration for this module.
   *
   * @var \Drupal\Core\Config\Config
   */
  protected Config $config;

  /**
   * @param array $configuration
   * @param $plugin_id
   * @param $plugin_definition
   * @param \Drupal\Core\Logger\LoggerChannelFactoryInterface $logger_factory
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   * @param \Drupal\Core\Template\TwigEnvironment $twig
   * @param \Drupal\Core\Datetime\DateFormatterInterface $date_formatter
   * @param \Drupal\notification_system_dispatch_webpush\WebPushClient $web_push_client
   * @param \Drupal\Core\File\FileSystemInterface $file_system
   * @param \Drupal\Core\Asset\LibraryDiscoveryInterface $library_discovery
   * @param \Drupal\Core\Render\ElementInfoManagerInterface $element_info_manager
   */
  public function __construct(array $configuration,
                              $plugin_id,
                              $plugin_definition,
                              LoggerChannelFactoryInterface $logger_factory,
                              ConfigFactoryInterface $config_factory,
                              TwigEnvironment $twig,
                              DateFormatterInterface $date_formatter,
                              WebPushClient $web_push_client,
                              FileSystemInterface $file_system,
                              LibraryDiscoveryInterface $library_discovery,
                              ElementInfoManagerInterface $element_info_manager) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);

    $this->twig = $twig;
    $this->dateFormatter = $date_formatter;
    $this->webPushClient = $web_push_client;
    $this->fileSystem = $file_system;
    $this->libraryDiscovery = $library_discovery;
    $this->elementInfoManager = $element_info_manager;

    $this->logger = $logger_factory->get('notification_system_dispatch_webpush');
    $this->config = $config_factory->getEditable('notification_system_dispatch_webpush.settings');
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('logger.factory'),
      $container->get('config.factory'),
      $container->get('twig'),
      $container->get('date.formatter'),
      $container->get('notification_system_dispatch_webpush'),
      $container->get('file_system'),
      $container->get('library.discovery'),
      $container->get('plugin.manager.element_info')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function settingsForm() {
    $form = parent::settingsForm();

    $form['help_text'] = [
      '#type' => 'markup',
      '#markup' => '
        <p>You can use twig to generate the notification title and body, which allows you to use if statements and for loops for the case that multiple notifications will be sent at once (notification bundling)</p>
        <p><strong>Variables:</strong></p>
        <ul>
          <li><em>notifications</em> - A list of notifications.</li>
          <ul>
            <li><em>title</em> - The title of the notification.</li>
            <li><em>body</em> - The body of the notification.</li>
            <li><em>timestamp</em> - The date and time when the notification was created. Uses the short date format.</li>
            <li><em>link</em> - A link to the notification. When clicking, the notification will be marked as read.</li>
            <li><em>direct_link</em> - A direct link to the notification without marking it as read.</li>
          </ul>
        </ul>
      ',
    ];

    $form['subject_template'] = [
      '#type' => 'textarea',
      '#title' => $this->t('Title Template'),
      '#default_value' => $this->config->get('subject_template'),
      '#rows' => 8,
      '#required' => TRUE,
    ];

    $form['body_template'] = [
      '#type' => 'textarea',
      '#title' => $this->t('Body Template'),
      '#default_value' => $this->config->get('body_template'),
      '#rows' => 20,
      '#required' => TRUE,
    ];



    $form['body_max_length'] = [
      '#type' => 'number',
      '#title' => $this->t('Body Max length (characters)'),
      '#default_value' => $this->config->get('body_max_length') ?: 200,
      '#required' => TRUE,
    ];

    $codeExample = new FormattableMarkup("<code><pre>openssl ecparam -genkey -name prime256v1 -out private.pem
openssl ec -in private.pem -pubout -outform DER|tail -c 65|base64|tr -d '=' |tr '/+' '_-' >> public.key
openssl ec -in private.pem -outform DER|tail -c +8|head -c 32|base64|tr -d '=' |tr '/+' '_-' >> private.key</pre></code>", []);

    $form['vapid_keys'] = [
      '#type' => 'fieldset',
      '#title' => $this->t('VAPID Keys'),
      '#description' => $this->t('Generate a key-pair for <a href="@url"
        target="_blank">Voluntary Application Server Identification (VAPID) for
        Web Push.</a><br>@codeExample', [
          '@url' => 'https://tools.ietf.org/id/draft-ietf-webpush-vapid-03.html',
          '@codeExample' => $codeExample,
        ]
      ),
    ];

    $form['vapid_keys']['vapid_public_key'] = [
      '#type' => 'textfield',
      '#title' => $this->t('VAPID Public Key'),
      '#default_value' => $this->config->get('vapid_public_key'),
      '#required' => TRUE,
    ];

    $form['vapid_keys']['vapid_private_key'] = [
      '#type' => 'textfield',
      '#title' => $this->t('VAPID Private Key'),
      '#default_value' => $this->config->get('vapid_private_key'),
      '#required' => TRUE,
    ];

    $form['badge'] = [
      '#type' => 'fieldset',
      '#title' => $this->t('Icon'),
      '#description' => $this->t('Specify an icon that represents your app.<br>
        It is recommended to be 128x128 px and a png file.<br>
        Also note that it may be converted to a monochrome style.'),
    ];

    $form['badge']['badge_path'] = [
      '#type' => 'textfield',
      '#title' => t('Path to custom icon'),
      '#default_value' => $this->config->get('badge_path'),
    ];

    $form['badge']['badge_upload'] = [
      '#type' => 'managed_file',
      '#title' => t('Upload icon image'),
      '#description' => t("If you don't have direct file access to the server, use this field to upload your icon."),
      '#upload_location' => 'public://webpush',
      '#upload_validators' => [
        'file_validate_is_image' => [],
      ],
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function settingsFormSubmit(array $values) {
    if (is_array($values['badge']['badge_upload']) && count($values['badge']['badge_upload']) > 0) {
      $file = File::load($values['badge']['badge_upload'][0]);
      if ($file instanceof File) {
        $file->setPermanent();
        $file->save();

        $values['badge']['badge_path'] = $file->getFileUri();
      }
    }

    // If vapid_public_key has been changed,
    // then need to rebuild site libraries and elements.
    if ($this->config->get('vapid_public_key') !== $values['vapid_keys']['vapid_public_key']) {
      $this->libraryDiscovery->clearCachedDefinitions();
      $this->elementInfoManager->clearCachedDefinitions();
    }

    $this->config
      ->set('subject_template', $values['subject_template'])
      ->set('body_template', $values['body_template'])
      ->set('body_max_length', $values['body_max_length'])
      ->set('vapid_public_key', $values['vapid_keys']['vapid_public_key'])
      ->set('vapid_private_key', $values['vapid_keys']['vapid_private_key'])
      ->set('badge_path', $values['badge']['badge_path'])
      ->save();

    parent::settingsFormSubmit($values);
  }

  /**
   * {@inheritdoc}
   */
  public function dispatch(UserInterface $user, array $notifications) {
    $vars = $this->getVars($user, $notifications);


    // Build push notification.
    $notification = new WebPushNotification($vars['subject']);
    $notification->setBody($vars['body']);
    $notification->setLanguage($user->getPreferredLangcode());


    // Build link. When a notification bundle is sent, link to the frontpage.
    $link = NULL;

    if (count($notifications) > 1) {
      $link = new Url('<front>');
      $link->setAbsolute(TRUE);
    }
    else {
      $link = $notifications[0]->getLink();

      if ($link) {
        $link->setAbsolute(TRUE);
      }
    }

    if ($link) {
      $notification->setData(new WebPushData([
        'open_url' => $link->toString(),
      ]));
    }


    // Set badge.
    if ($this->config->get('badge_path') && strlen($this->config->get('badge_path')) > 0) {
      $badge_path = $this->config->get('badge_path');

      if (StreamWrapperManager::getScheme($badge_path)) {
        $notification->setBadge(file_create_url($badge_path));
      }
      else {
        $notification->setBadge($badge_path);
      }
    }


    // Send notification.
    try {
      $this->webPushClient->sendToUser($user->id(), $notification);
    }
    catch (\ErrorException $e) {
      $this->logger->warning('Error while sending the notifications to ' . $user->label() . ' (UID ' . $user->id() . ') via web push. ' . $e->getMessage());
    }
  }

  /**
   * Generates the subject and body of the notification.
   *
   * @param \Drupal\user\UserInterface $user
   *   The user who this notification is for.
   * @param \Drupal\notification_system\model\Notification[] $notifications
   *   An array of notifications to send.
   *
   * @return array
   *   - subject: string
   *   - body: string
   */
  protected function getVars(UserInterface $user, array $notifications) {
    $langcode = $user->getPreferredLangcode();

    $variables = [
      'notifications' => [],
    ];

    foreach ($notifications as $notification) {
      $direct_link = $notification->getLink();
      if ($direct_link) {
        $direct_link = $direct_link->setAbsolute(TRUE)->toString();
      }

      // Convert body to plain text.
      $bodyText = $notification->getBody();
      $bodyHtml = new Html2Text($bodyText, [
        'width' => 0,
        'links' => 'none',
      ]);
      $bodyText = $bodyHtml->getText();

      $notificationVariables = [
        'title' => $notification->getTitle(),
        'body' => $bodyText,
        'timestamp' => $this->dateFormatter->format($notification->getTimestamp(), 'medium', '', NULL, $langcode),
        'link' => 'http://example.com',
        'direct_link' => $direct_link,
      ];

      $variables['notifications'][] = $notificationVariables;
    }

    $subjectTemplate = '{% spaceless %}' . $this->config->get('subject_template') . '{% endspaceless %}';
    $bodyTemplate = $this->config->get('body_template');

    $subject = htmlspecialchars_decode($this->twig->renderInline($subjectTemplate, $variables));
    $body = htmlspecialchars_decode($this->twig->renderInline($bodyTemplate, $variables));


    // Limit length of body.
    $maxLength = $this->config->get('body_max_length') ?: 200;
    if (strlen($body) > $maxLength) {
      $body = substr($body, 0, $maxLength) . ' ...';
    }

    return [
      'subject' => $subject,
      'body' => $body,
    ];
  }

}
